from ipywidgets import VBox

def crop(image,
         slice_number: int = None,
         axis: int = 0,
         continuous_update: bool = True,
         slider_text: str = "Slice",
         axis_names=None,
         zoom_factor: float = 1.0,
         zoom_spline_order: int = 0,
         colormap:str = None,
         display_min:float = None,
         display_max:float = None):
    """
    Allows cropping an image along all axes.

    Parameters
    ----------
    image : image
        2D or 3D image to be cropped
    slice_number : int, optional
        Slice-position in the stack to be shown (default: center plane)
    axis : int, optional
        Axis along which we slice the shown stack (default: 0)
    continuous_update : bool, optional
        Update the image while dragging the sliders, default: False
    slider_text: str, optional
        Text to be shown next to the slider
    axis_names: list of str, optional
        Names of the axes. If not given, the default names Z,Y,X are used.
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities

    """

    if 'cupy.ndarray' in str(type(image)):
        image = image.get()

    return _Cropper(image,
                    slice_number=slice_number,
                    axis=axis,
                    continuous_update=continuous_update,
                    slider_text=slider_text,
                    axis_names=axis_names,
                    zoom_factor=zoom_factor,
                    zoom_spline_order=zoom_spline_order,
                    colormap=colormap,
                    display_min=display_min,
                    display_max=display_max)




class _Cropper(VBox):
    """
    See also
    --------
    crop()
    """
    def __init__(self,
                 image,
                 slice_number: int = None,
                 axis: int = 0,
                 continuous_update: bool = False,
                 slider_text: str = "Slice",
                 axis_names=None,
                 zoom_factor: float = 1.0,
                 zoom_spline_order: int = 0,
                 colormap:str = None,
                 display_min:float = None,
                 display_max:float = None):
        from ipywidgets import IntRangeSlider
        from ._slice_viewer import _SliceViewer
        from ._utilities import _no_resize

        self._image = image

        viewer = _SliceViewer(image,
                              slice_number,
                              axis,
                              0,
                              0,
                              continuous_update,
                              slider_text,
                              zoom_factor=zoom_factor,
                              zoom_spline_order=zoom_spline_order,
                              colormap=colormap,
                              display_min=display_min,
                              display_max=display_max
                              )

        if len(image.shape) < 2 or len(image.shape) > 3:
            raise RuntimeError("Number of image dimensions must be 2 or 3, but is", len(image.shape))

        if axis_names is None:
            if len(image.shape) == 2:
                axis_names = ["Y", "X"]
            elif len(image.shape) == 3:
                axis_names = ["Z", "Y", "X"]

        self._viewer = viewer
        view = viewer.view
        slice_slider = viewer.slice_slider

        self._range_sliders = []
        for dim in range(len(image.shape)):
            min_ = 0
            max_ = image.shape[dim]

            range_slider = IntRangeSlider(
                value=[min_, max_],
                min=min_,
                max=max_,
                description=axis_names[dim],
                continuous_update=continuous_update,
            )
            range_slider.observe(self.update)
            self._range_sliders.append(range_slider)

        widgets = []
        widgets = widgets + self._range_sliders
        widgets.append(_no_resize(view))
        if len(image.shape) > 2:
            widgets.append(slice_slider)

        super(_Cropper, self).__init__(widgets)

    def update(self, event=None):
        self._viewer.image = self._image[self.range]
        self._viewer.slice_slider.max = self._viewer.image.shape[0] - 1
        self._viewer.update(None)

    @property
    def range(self):
        """
        Returns a tuple of Python slice objects that can be used to crop the image.
        """
        return tuple([slice(r.value[0], r.value[1], 1) for r in self._range_sliders])

    @range.setter
    def range(self, slices):
        """
        Sets the range of the sliders to the given Python slice objects.
        The step of the slices is ignored.
        """
        for a_slice, a_range_slider in zip(slices, self._range_sliders):
            a_range_slider.value = [a_slice.start, a_slice.stop]

    def crop(self):
        """
        Returns the cropped image.
        """
        from stackview import insight
        return insight(self._image[self.range])




def animate(timelapse, filename:str=None, overwrite_file:bool=True, frame_delay_ms:int=150, num_loops:int=1000, colormap=None, display_min=None, display_max=None, zoom_factor:float=1.0):
    """
    Create an animated GIF from a list of 2D images and return it as Markdown object, that can be shown in Jupyter notebooks.
    RGB images are supported as well, but no 3D image stacks.

    Note: It is recommended to normalize the timelapse to be integers in the range between 0 to 255.

    Parameters
    ----------
    timelapse: np.ndarray
        2D image or 3D image stack
    filename: str, optional
        Name of the file where the animation will be saved
    overwrite_file: bool, optional
        Overwrite the file if it already exists. Default: True
    frame_delay_ms: int, optional
        Delay between frames in milliseconds
    num_loops: int, optional
        Number of loops in the animation
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities
    """
    import os
    import imageio
    from IPython.display import HTML
    import numpy as np
    import warnings
    from stackview._image_widget import _img_to_rgb
    from ._utilities import numpy_to_gif_bytestream, _gif_to_html

    if 0 <= timelapse.min() <= 1 and 0 <= timelapse.max() <= 1:
        warnings.warn("The timelapse has a small intensity range between 0 and 1. Consider normalizing it to the range between 0 and 255.")
    if timelapse.min() < 0 or timelapse.max() > 255:
        warnings.warn("The timelapse has an intensity range exceeding 0..255. Consider normalizing it to the range between 0 and 255.")

    image_rgb = np.asarray([_img_to_rgb(i, colormap=colormap, display_min=display_min, display_max=display_max) for i in timelapse]).astype(np.uint8)

    if filename is not None:
        if not filename.endswith(".gif"):
            filename += ".gif"
        if not overwrite_file:
            i = 0
            original_filename = filename
            while os.path.exists(filename):
                i += 1
                filename = original_filename.replace(".gif", f"_{i:02}.gif")

        with imageio.get_writer(filename, mode='I', duration=frame_delay_ms, loop=num_loops) as writer:
            for frame in image_rgb:
                writer.append_data(frame)


    if np.prod(image_rgb.shape) > 1024 * 1024 * 10:
        warnings.warn("The image is quite large (> 10 MByte) and might not be properly shown in the notebook when rendered over the internet. Consider subsampling or cropping the image for visualization purposes.")

    bytestream = numpy_to_gif_bytestream(image_rgb, frame_delay_ms=frame_delay_ms, num_loops=num_loops)
    return HTML(_gif_to_html(bytestream, width=int(image_rgb.shape[-2] * zoom_factor)))

def animate_curtain(timelapse, timelapse_curtain,
                    axis: int = 0,
                    alpha: float = 1,
                    num_steps:int = 20,
                    zoom_factor: float = 1,
                    colormap=None, display_min=None, display_max=None,
                    curtain_colormap:str = None,
                    curtain_display_min:float = None,
                    curtain_display_max:float = None,
                    filename:str=None, overwrite_file:bool=False, frame_delay_ms:int=150, num_loops:int=1000):
    """
    Render a curtain showing two 2D images. RGB images are supported as well, but no 3D image stacks.

    Parameters
    ----------
    timelapse: np.ndarray
        2D image or 3D image stack
    timelapse_curtain: np.ndarray
        2D image or 3D image stack
    axis: int, optional
        Axis in case we are slicing a stack
    alpha: float, optional
        sets the transparency of the curtain
    num_steps: int, optional
        number of steps in the animation, half of them the curtain will go left-right, the other half right-left
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities
    curtain_colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    curtain_display_min: float, optional
        Lower bound of properly shown intensities
    curtain_display_max: float, optional
        Upper bound of properly shown intensities
    filename: str, optional
        Name of the file where the animation will be saved
    overwrite_file: bool, optional
        Overwrite the file if it already exists
    frame_delay_ms: int, optional
        Delay between frames in milliseconds
    num_loops: int, optional
        Number of loops in the animation

    """
    import numpy as np
    from ._image_widget import _img_to_rgb

    max_size = timelapse.shape[1]

    images = []
    steps = list(range(0, max_size + 1, int(max_size / num_steps * 2)))
    steps = steps + steps[::-1]

    for slider_value in steps:
        if len(timelapse.shape) < 3 or (len(timelapse.shape) == 3 and timelapse.shape[-1] == 3):
            image_slice = _img_to_rgb(timelapse.copy(), colormap=colormap, display_min=display_min, display_max=display_max)
            image_slice_curtain = _img_to_rgb(timelapse_curtain, colormap=curtain_colormap, display_min=curtain_display_min,
                                              display_max=curtain_display_max)
        else:
            image_slice = _img_to_rgb(np.take(timelapse, slider_value, axis=axis), colormap=colormap,
                                      display_min=display_min, display_max=display_max)
            image_slice_curtain = _img_to_rgb(np.take(timelapse_curtain, slider_value, axis=axis),
                                              colormap=curtain_colormap, display_min=curtain_display_min,
                                              display_max=curtain_display_max)
        image_slice[:, slider_value:] = (1 - alpha) * image_slice[:, slider_value:] + \
                                                alpha * image_slice_curtain[:, slider_value:]

        images.append(image_slice)

    return animate(np.asarray(images), filename=filename, overwrite_file=overwrite_file, frame_delay_ms=frame_delay_ms, num_loops=num_loops, zoom_factor=zoom_factor)




def picker(
        image,
        slice_number: int = None,
        display_width: int = None,
        display_height: int = None,
        continuous_update: bool = True,
        slider_text: str = "Slice",
        zoom_factor:float = 1.0,
        zoom_spline_order:int = 0,
        colormap:str = None,
        display_min:float = None,
        display_max:float = None
):
    """Shows an image with a slider to go through a stack plus a label with the current mouse position and intensity at that position.

    Parameters
    ----------
    image : image
        Image shown
    slice_number : int, optional
        Slice-position in the stack
    display_width : int, optional
        This parameter is obsolete. Use zoom_factor instead
    display_height : int, optional
        This parameter is obsolete. Use zoom_factor instead
    continuous_update : bool, optional
        Update the image while dragging the mouse, default: False
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities

    Returns
    -------
    An ipywidget with an image display, a slider and a label showing mouse position and intensity.
    """

    import ipywidgets
    from ._slice_viewer import _SliceViewer
    from ._utilities import _no_resize

    if 'cupy.ndarray' in str(type(image)):
        image = image.get()

    viewer = _SliceViewer(image,
                          slice_number=slice_number,
                          continuous_update=continuous_update,
                          slider_text=slider_text,
                          zoom_factor=zoom_factor,
                          zoom_spline_order=zoom_spline_order,
                          colormap=colormap,
                          display_min=display_min,
                          display_max=display_max
                          )
    view = viewer.view
    slice_slider = viewer.slice_slider
    label = ipywidgets.Label("[]:")

    from ipyevents import Event
    event_handler = Event(source=view, watched_events=['mousemove'])

    def update_display(event):
        relative_position_x = event['relativeX'] / zoom_factor
        relative_position_y = event['relativeY'] / zoom_factor
        absolute_position_x = int(relative_position_x)
        absolute_position_y = int(relative_position_y)

        if len(image.shape) > 2:
            absolute_position_z = slice_slider.value
            intensity = image[absolute_position_z, absolute_position_y, absolute_position_x]
            label.value = "[z=" + str(absolute_position_z) + ", y=" + str(absolute_position_y) + ", x=" + str(
                absolute_position_x) + "] = " + str(intensity)
        else:
            intensity = image[absolute_position_y, absolute_position_x]
            label.value = "[y=" + str(absolute_position_y) + ", x=" + str(absolute_position_x) + "] = " + str(intensity)

    event_handler.on_dom_event(update_display)

    result = ipywidgets.VBox([_no_resize(view), slice_slider, label], stretch=False)
    result.update = update_display
    return result



import numpy as np
from typing import Callable
from functools import wraps
from toolz import curry

def insight(image, library_name=None, help_url=None):
    """
    Converts a numpy-array-like image to numpy-compatible array with a convenient display in Jupyter notebooks
    including shape, min/max intensity, histogram and viewing 32-bit and 64-bit integer images as coloured labels.
    """
    return StackViewNDArray(image, library_name, help_url)

@curry
def jupyter_displayable_output(
        function: Callable,
        library_name:str = None,
        help_url:str = None
) -> Callable:
    """Wraps a given function so that it outputs a nice image view in jupyter notebooks
    The view will contain a link to the specified library to guide users to read more.
    """
    @wraps(function)
    def worker_function(*args, **kwargs):
        # call the decorated function
        result = function(*args, **kwargs)

        # Attach _repr_html_ function
        result = StackViewNDArray(result, library_name, help_url)

        return result

    return worker_function


class StackViewNDArray(np.ndarray):

    def __new__(cls, input_array, library_name=None, help_url=None):
        if 'cupy.ndarray' in str(type(input_array)):
            input_array = input_array.get()
        obj = np.asarray(input_array).view(cls)
        obj.library_name = library_name
        obj.help_url = help_url
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.library_name = getattr(obj, 'library_name', None)
        self.help_url = getattr(obj, 'help_url', None)

    def _repr_html_(self):
        """HTML representation of the image object for IPython.
                Returns
                -------
                HTML text with the image and some properties.
                """
        if len(self.shape) < 2:
            return str(self)

        import numpy as np
        size_in_pixels = np.prod(self.shape)
        size_in_bytes = size_in_pixels * self.dtype.itemsize

        from ._image_widget import _is_label_image
        labels = _is_label_image(self)

        import matplotlib.pyplot as plt
        _imshow(self,
                labels=labels,
                continue_drawing=True,
                colorbar=not labels)
        image = _png_to_html(_plt_to_png())

        if size_in_bytes > 1024:
            size_in_bytes = size_in_bytes / 1024
            if size_in_bytes > 1024:
                size_in_bytes = size_in_bytes / 1024
                if size_in_bytes > 1024:
                    size_in_bytes = size_in_bytes / 1024
                    size = "{:.1f}".format(size_in_bytes) + " GB"
                else:
                    size = "{:.1f}".format(size_in_bytes) + " MB"
            else:
                size = "{:.1f}".format(size_in_bytes) + " kB"
        else:
            size = "{:.1f}".format(size_in_bytes) + " B"

        histogram = ""

        if size_in_bytes < 100 * 1024 * 1024:
            if not labels:
                import numpy as np

                num_bins = 32
                h, _ = np.histogram(self, bins=num_bins)

                plt.figure(figsize=(1.8, 1.2))
                plt.bar(range(0, len(h)), h)

                # hide axis text
                # https://stackoverflow.com/questions/2176424/hiding-axis-text-in-matplotlib-plots
                # https://pythonguides.com/matplotlib-remove-tick-labels
                frame1 = plt.gca()
                frame1.axes.xaxis.set_ticklabels([])
                frame1.axes.yaxis.set_ticklabels([])
                plt.tick_params(left=False, bottom=False)

                histogram = _png_to_html(_plt_to_png())

            min_max = "<tr><td>min</td><td>" + str(self.min()) + "</td></tr>" + \
                      "<tr><td>max</td><td>" + str(self.max()) + "</td></tr>"

        else:

            min_max = ""

        help_text = ""
        if self.library_name is not None and len(self.library_name) > 0:
            self.library_name = self.library_name + " made "
            if self.help_url is not None:
                help_text = "<b><a href=\"" + self.help_url + "\" target=\"_blank\">" + self.library_name + "</a>image</b><br/>"

        all = [
            "<table>",
            "<tr>",
            "<td>",
            image,
            "</td>",
            "<td style=\"text-align: center; vertical-align: top;\">",
            help_text,
            "<table>",
            "<tr><td>shape</td><td>" + str(self.shape).replace(" ", "&nbsp;") + "</td></tr>",
            "<tr><td>dtype</td><td>" + str(self.dtype) + "</td></tr>",
            "<tr><td>size</td><td>" + size + "</td></tr>",
            min_max,
            "</table>",
            histogram,
            "</td>",
            "</tr>",
            "</table>",
        ]

        return "\n".join(all)


def _png_to_html(png):
    import base64
    url = 'data:image/png;base64,' + base64.b64encode(png).decode('utf-8')
    return f'<img src="{url}"></img>'


# adapted from https://github.com/napari/napari/blob/d6bc683b019c4a3a3c6e936526e29bbd59cca2f4/napari/utils/notebook_display.py#L54-L73
def _plt_to_png():
    """PNG representation of the image object for IPython.
    Returns
    -------
    In memory binary stream containing a PNG matplotlib image.
    """
    import matplotlib.pyplot as plt
    from io import BytesIO

    with BytesIO() as file_obj:
        plt.savefig(file_obj, format='png')
        plt.close() # supress plot output
        file_obj.seek(0)
        png = file_obj.read()
    return png


def _imshow(image, title: str = None, labels: bool = False, min_display_intensity: float = None,
            max_display_intensity: float = None, plot=None, colorbar: bool = False, colormap=None,
            alpha: float = None, continue_drawing: bool = False):
    """Visualize an image, e.g. in Jupyter notebooks.

    Parameters
    ----------
    image: np.ndarray
        numpy or OpenCL-backed image to visualize
    title: str
        Obsolete (kept for ImageJ-compatibility)
    labels: bool
        True: integer labels will be visualized with colors
        False: Specified or default colormap will be used to display intensities.
    min_display_intensity: float
        lower limit for display range
    max_display_intensity: float
        upper limit for display range
    color_map: str
        deprecated, use colormap instead
    plot: matplotlib axis
        Plot object where the image should be shown. Useful for putting multiple images in subfigures.
    colorbar: bool
        True puts a colorbar next to the image. Will not work with label images and when visualizing multiple
        images (continue_drawing=True).
    colormap: str or matplotlib colormap
    alpha: float
        alpha blending value
    continue_drawing: float
        True: the next shown image can be visualized on top of the current one, e.g. with alpha = 0.5
    """
    import numpy as np

    if len(image.shape) == 3 and image.shape[2] == 3: # RGB image
        import matplotlib.pyplot as plt
        plt.imshow(image, vmin=min_display_intensity, vmax=max_display_intensity,
                   interpolation='nearest', alpha=alpha)
        if not continue_drawing:
            plt.show()
        return

    if len(image.shape) == 3:
        image = np.asarray(image).max(axis=0)

    image = np.asarray(image)
    if len(image.shape) == 1:
        image = image[np.newaxis]

    if colormap is None:
        colormap = "Greys_r"

    cmap = colormap
    if labels:
        import matplotlib
        import numpy as np

        if not hasattr(_imshow, "labels_cmap"):
            from ._image_widget import _labels_lut
            _imshow.labels_cmap = matplotlib.colors.ListedColormap(_labels_lut())
        cmap = _imshow.labels_cmap

        if min_display_intensity is None:
            min_display_intensity = 0
        if max_display_intensity is None:
            max_display_intensity = 65536

    if plot is None:
        import matplotlib.pyplot as plt
        plt.imshow(image, cmap=cmap, vmin=min_display_intensity, vmax=max_display_intensity,
                   interpolation='nearest', alpha=alpha)
        if colorbar:
            plt.colorbar()
        if not continue_drawing:
            plt.show()
    else:
        plot.imshow(image, cmap=cmap, vmin=min_display_intensity, vmax=max_display_intensity,
                    interpolation='nearest', alpha=alpha)
        if colorbar:
            plot.colorbar()


__version__ = "0.7.10"

from ._static_view import jupyter_displayable_output, insight
from ._utilities import merge_rgb
from ._context import nop
from ._crop import crop
from ._slice_viewer import _SliceViewer
from ._annotate import annotate
from ._utilities import _no_resize
from ._interact import interact
from ._slice import slice
from ._curtain import curtain
from ._orthogonal import orthogonal
from ._side_by_side import side_by_side
from ._picker import picker
from ._assist import assist
from ._switch import switch
from ._colormaps import create_colormap
from ._imshow import imshow
from ._animate import animate, animate_curtain
from ._display_range import display_range





import warnings


def draw_circle(x, y, z, axis, radius, value, image):
    from skimage.draw import disk

    # draw circle
    rr, cc = disk((y, x), radius)
    if z is not None:
        if axis == 0:
            image_slice = image[z]
        elif axis == 1:
            image_slice = image[:, z]
        elif axis == 2:
            image_slice = image[:, :, z]
        image_slice[rr, cc] = value
    else:
        image[rr, cc] = value


def annotate(
        image,
        labels,
        slice_number: int = None,
        alpha: float = 0.5,
        axis: int = 0,
        continuous_update: bool = True,
        slider_text: str = "Slice",
        zoom_factor: float = 1.0,
        zoom_spline_order: int = 0,
        colormap:str = None,
        display_min:float = None,
        display_max:float = None
):
    """Shows an image with a slider to go through a stack plus a label with the current mouse position and intensity at that position.

    Parameters
    ----------
    image : image
        Image shown
    labels: label image
        Labels which can be manually modified to draw annotations
    slice_number : int, optional
        Slice-position in the stack
    axis: int, optional
        Axis where the slice-position can be modified interactively
    alpha : float, optional
        Alpha blending value for the labels on top of the image
    continuous_update : bool, optional
        Update the image while dragging the mouse on the slider, default: False
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities

    Returns
    -------
    An ipywidget with an image display, a slider and a label showing mouse position and intensity.
    """
    from ._utilities import _no_resize
    from ._image_widget import _is_label_image, _img_to_rgb
    import ipywidgets
    from ._image_widget import ImageWidget
    import numpy as np
    from ipyevents import Event
    from ._uint_field import UIntField

    if 'cupy.ndarray' in str(type(image)):
        image = image.get()

    if 'cupy.ndarray' in str(type(labels)):
        labels = labels.get()

    if not _is_label_image(labels):
        warnings.warn("Labels are not an image of type uint32 or uint64. Consider converting to this type for best performance.")

    if slice_number is None:
        slice_number = int(image.shape[axis] / 2)

    if len(image.shape) <= 2:
        slice_image = image
    else:
        slice_image = np.take(image, slice_number, axis=axis)

    # Image view
    view = ImageWidget(slice_image, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order)

    # setup user interface for changing the slice
    slice_slider = None
    if len(image.shape) > 2:
        slice_slider = ipywidgets.IntSlider(
            value=slice_number,
            min=0,
            max=image.shape[axis] - 1,
            continuous_update=continuous_update,
            description=slider_text,
        )

    # event handler when the user changed the slider:
    def update_display(event=None):
        if slice_slider is not None:
            z = slice_slider.value
            slice_image1 = np.take(image, z, axis=axis)
            slice_image2 = np.take(labels, z, axis=axis)
        else:
            slice_image1 = image
            slice_image2 = labels

        if not _is_label_image(slice_image2):
            # necessary to display the labels in colour correctly
            slice_image2 = slice_image2.astype(np.uint64)

        rgb_image1 = _img_to_rgb(slice_image1, colormap=colormap, display_min=display_min, display_max=display_max)
        rgb_image2 = _img_to_rgb(slice_image2, colormap=colormap, display_min=display_min, display_max=display_max)

        factor1 = 1.0 - alpha
        factor2 = alpha

        rgb_mix = factor1 * rgb_image1 + factor2 * rgb_image2

        view.data = rgb_mix

    # user interface for drawing
    label_id_slider = UIntField(1)
    radius_slider = UIntField(2)
    radius_eraser_slider = UIntField(4)
    tool_box = ipywidgets.VBox([
        ipywidgets.Label("Label ID:"),
        label_id_slider,
        ipywidgets.Label("Draw radius:"),
        radius_slider,
        ipywidgets.Label("Eraser radius:"),
        radius_eraser_slider
    ])

    event_handler = Event(source=view, watched_events=['mousemove'])

    former_drawn_position = [None, 0, 0, 0]

    # event handler for when something was drawn
    def update_display_while_drawing(event):
        if event['buttons'] == 0:
            former_drawn_position[0] = None
            return

        # 'altKey' makes use of the eraser
        label_id_to_draw = label_id_slider.value
        radius = radius_slider.value
        if event['altKey']:
            label_id_to_draw = 0
            radius = radius_eraser_slider.value

        # get position from event
        relative_position_x = event['relativeX'] / zoom_factor
        relative_position_y = event['relativeY'] / zoom_factor
        absolute_position_x = int(relative_position_x)
        absolute_position_y = int(relative_position_y)

        # differentiate 3D/2D drawing
        if slice_slider is not None:
            absolute_position_z = slice_slider.value
            position = [absolute_position_x, absolute_position_y, absolute_position_z]
        else:
            absolute_position_z = None
            position = [absolute_position_x, absolute_position_y]

        # compare position and label with last known postion. If equal, don't update / redraw
        position.append(label_id_to_draw)
        if np.array_equal(former_drawn_position, position):
            return

        # draw one circle
        draw_circle(absolute_position_x,
                    absolute_position_y,
                    absolute_position_z,
                    axis,
                    radius,
                    label_id_to_draw,
                    labels)

        # draw circles along a line we've beend drawing
        if former_drawn_position[0] is not None:
            distance = np.linalg.norm(np.asarray(position[0:2]) - np.asarray(former_drawn_position[0:2]))
            for i in range(int(distance)):
                relative_position = i / int(distance)
                position_to_draw = relative_position * np.asarray(former_drawn_position[0:2]) + \
                                   (1.0 - relative_position) * np.asarray(position[0:2])
                draw_circle(position_to_draw[0],
                            position_to_draw[1],
                            absolute_position_z,
                            axis,
                            radius,
                            label_id_to_draw,
                            labels)

        # store position
        for i in range(4):
            if i < len(position):
                former_drawn_position[i] = position[i]
        update_display()

    # draw everything once
    update_display()

    # connect events
    event_handler.on_dom_event(update_display_while_drawing)

    if slice_slider is not None:
        # connect user interface with event
        slice_slider.observe(update_display)

        result = ipywidgets.VBox([
            ipywidgets.HBox([_no_resize(view), tool_box]),
            slice_slider
        ])
    else:
        result = ipywidgets.VBox([
            ipywidgets.HBox([_no_resize(view), tool_box]),
        ])
    result.update = update_display
    return result



def assist(what: str = "",
           context:dict = None,
           continuous_update: bool = True,
           zoom_factor: float = 1.0,
           zoom_spline_order: int = 0,
           ):
    """
    Searches provided context (e.g. globals()) for functions that have images as parameters and provide
    a pulldown of these functions. Once the user selected a function from this pulldown, a user-interface
    is created for the function's parameters.

    Note
    ----
    There is currently a technical limitation to about 250 list entries. This is due to a maximum number of GUI
    elements. Thus, the search term `what` should be as specific as possible to limit the number of results.
    If you receive a JavaScript error or the progress bar gets stuck, try to be more specific.

    See Also
    --------
    stackview.interact()

    Parameters
    ----------
    what : str
        The task to be assisted, e.g. "denoise", "filter", "morphology", "transform", "segment", "map", "mesh"
    context : dict
        A dictionary of (name:image/function/module), allows showing pulldowns of available functions and images, e.g.: globals()
    continuous_update : bool, optional
        Update the image while dragging the mouse, default: False
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    """
    import ipywidgets
    from ._context import Context
    from ._utilities import _no_resize
    from ._slice_viewer import _SliceViewer
    from ._utilities import logo
    from ._interact import interact

    # search is case-insensitive
    what = what.lower()

    if not isinstance(context, Context):
        context = Context(context)

    # GUI: The image viewer for all function widgets is this one:
    viewer = _SliceViewer(logo, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order, continuous_update=continuous_update)

    # search compatible functions and make widgets for them
    available_widgets = {}
    search_index = {}
    count = 0
    for index, (name, func) in enumerate(context._functions.items()):
        if hasattr(func, "categories"):
            if func.categories is None or "in assistant" not in func.categories:
                continue
        try:
            widget = interact((func, name), context=context, viewer=viewer, continuous_update=continuous_update, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order)
            search_text = (func.__name__ + str(func.__doc__)).lower()
        except:
            del widget
            continue
        if search_string(what, search_text):
            count = count + 1
            search_index[search_text] = name
            available_widgets[name] = widget
            import inspect
            sig = inspect.signature(func)
        else:
            del widget

    # GUI: documentation output
    documentation_output = ipywidgets.HTML("")
    documentation_output.style.font_family = "Courier"

    # GUI: select operation
    operation_pulldown = ipywidgets.Dropdown(
        options=list(available_widgets.keys()),
        description="Operation"
    )
    def hide_all():
        """Hide all widgets"""
        for name, wi in available_widgets.items():
            wi.layout.display = 'none'
    def on_change(event):
        """Show the selected function widget"""
        if event['type'] == 'change' and event['name'] == 'value':
            func_name = event['new']
            hide_all()
            wi = available_widgets[func_name]
            wi.layout.display = 'block'
            documentation_output.value = text2html(context._functions[func_name].__doc__)
    operation_pulldown.observe(on_change)

    # GUI: Search field
    search_field = ipywidgets.Text(
        value='',
        placeholder='Type here to search',
        description='Search'
    )
    def search(event=None):
        """Update the operation pulldown based on search text"""
        search_text = str(search_field.value).lower()
        if len(search_text) < 1:
            operation_pulldown.options = list(available_widgets.keys())
        result = []
        for docstring, func_name in search_index.items():
            if search_string(search_text, docstring):
                result.append(func_name)
        operation_pulldown.options = result
    search_field.observe(search)

    # combine all widgets
    widgets_to_show = [search_field, operation_pulldown]
    for name, w in available_widgets.items():
        widgets_to_show.append(w)
    widgets_to_show.append(_no_resize(viewer.view))
    widgets_to_show.append(viewer.slice_slider)
    widgets_to_show.append(documentation_output)

    hide_all()

    return ipywidgets.VBox(widgets_to_show)


def text2html(text):
    html = "<pre>" + str(text) + "</pre>"
    return html

def search_string(what, where):
    return what in where





def list_bia_bob_plugins():
    """List of function hints for bia_bob"""
    return """    * Displays an image with a slider and label showing mouse position and intensity.
    stackview.annotate(image, labels)
    
    * Allows cropping an image along all axes.
    stackview.crop(image)
    
    * Showing an image stored in variable `image` and a segmented image stored in variable `labels` on top. Also works with two images or two label images.
    stackview.curtain(image, labels, alpha: float = 1)
    
    * Display an image stored in a variable `image` (this also works with label images). Prefer stackview.insight over matplotlib.pyplot.imshow!
    stackview.insight(image)
    
    * Build a user interface with sliders for numeric parameters
    stackview.interact(func, image)
    
    * Display an image and allows slicing in three dimensions (Z, Y, and X).
    stackview.orthogonal(image)
    
    * Display an image with a slider and label showing the mouse position and intensity.
    stackview.picker(image)
    
    * Display two images side by side, with an additional overlay view of their overlap
    stackview.side_by_side(image1, image2)
    
    * Display an image with a slider to navigate through a stack.
    stackview.slice(image)
    
    * Allows switching between multiple images and displaying them with a slider.
    stackview.switch(images:list)
    
    """



import ipywidgets

class UIntField(ipywidgets.HBox):
    """
    A text field where users can select a positive integer number from.
    Similar to ipywidgets.IntText
    """

    def __init__(self, value):
        self._value = value
        self._label = ipywidgets.Label(str(value))
        layout = ipywidgets.Layout(min_width='10px', max_width='30px')
        self._button_up = ipywidgets.Button(description="+", layout=layout)
        self._button_down = ipywidgets.Button(description="-", layout=layout)
        super().__init__([self._button_down, self._label, self._button_up])

        def increase(event=None):
            self.value = self.value + 1

        def decrease(event=None):
            self.value = self.value - 1

        self._button_up.on_click(increase)
        self._button_down.on_click(decrease)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if value >= 0:
            self._value = value
            self._label.value = str(value)


def switch(images,
           slice_number : int = None,
           axis : int = 0,
           display_width : int = None,
           display_height : int = None,
           continuous_update:bool = True,
           slider_text:str="Slice",
           zoom_factor:float = 1.0,
           zoom_spline_order:int = 0,
           colormap:str = None,
           display_min:float = None,
           display_max:float = None,
           toggleable:bool = False
):
    """Allows switching between multiple images.

    Parameters
    ----------
    images : list(image), or dict(str:image)
        Images to switch between. All images should have the same shape.
    slice_number : int, optional
        Slice-position in the stack
    axis : int, optional
        Axis in case we are slicing a stack
    display_width : int, optional
        This parameter is obsolete. Use zoom_factor instead
    display_height : int, optional
        This parameter is obsolete. Use zoom_factor instead
    continuous_update : bool, optional
        Update the image while dragging the mouse, default: False
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, or list of str optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float or list of float, optional
        Lower bound of properly shown intensities
    display_max: float of list of float, optional
        Upper bound of properly shown intensities
    toggleable: bool, optional
        If true, multiple images can be shown at the same time.
        In that case, it is recommended to pass a list as colormap as well.

    Returns
    -------
    An ipywidget with an image display and a slider.
    """
    from ._utilities import _no_resize
    from ._slice_viewer import _SliceViewer
    import ipywidgets
    from functools import partial
    from ._image_widget import _img_to_rgb
    import numpy as np

    if not isinstance(colormap, list):
        colormap = [colormap] * len(images)
    if not isinstance(display_min, list):
        display_min = [display_min] * len(images)
    if not isinstance(display_max, list):
        display_max = [display_max] * len(images)

    if isinstance(images, dict):
        names = list(images.keys())
        images = list(images.values())
        layout = None
    else:
        names = [str(i) for i in range(len(images))]
        layout = ipywidgets.Layout(min_width='10px', max_width='30px')

    images = [image.get() if 'cupy.ndarray' in str(type(image)) else image for image in images]

    viewer = _SliceViewer(images[0],
                          slice_number,
                          axis,
                          display_width,
                          display_height,
                          continuous_update,
                          slider_text,
                          zoom_factor=zoom_factor,
                          zoom_spline_order=zoom_spline_order,
                          colormap=colormap[0],
                          display_min=display_min[0],
                          display_max=display_max[0]
                          )
    view = viewer.view
    slice_slider = viewer.slice_slider

    buttons = []

    if toggleable:
        def display_(buttons, images, colormap, display_min, display_max):
            display_image = np.zeros(list(images[0].shape) + [3])
            for button, image, colormap_, display_min_, display_max_ in zip(buttons, images, colormap,
                                                                             display_min, display_max):
                if button.value:
                    if len(image.shape) == 3 and image.shape[-1] != 3:
                        display_image_to_add = np.asarray([_img_to_rgb(i, display_min=display_min_, display_max=display_max_, colormap=colormap_) for i in image])
                    else:
                        display_image_to_add = _img_to_rgb(image, display_min=display_min_, display_max=display_max_, colormap=colormap_)
                        
                    if display_image is None:
                        display_image = display_image_to_add
                    else:
                        display_image = display_image + display_image_to_add

            display_image = np.minimum(display_image, np.asarray([[255]]))

            viewer.view.colormap = None
            viewer.view.display_min = None
            viewer.view.display_max = None
            viewer.image = display_image
            viewer.update()

        display_function = partial(display_, buttons, images, colormap, display_min, display_max)
    else:
        def display_(image, colormap, display_min, display_max):
            viewer.view.colormap = colormap
            viewer.view.display_min = display_min
            viewer.view.display_max = display_max
            viewer.image = image
            viewer.update()

    for i, name, image, colormap_, display_min_, display_max_ in zip(range(len(names)), names, images, colormap, display_min, display_max):
        if toggleable:
            button = _make_toggle_button(name, layout, True, display_function)
        else:
            display_function = partial(display_, image, colormap_, display_min_, display_max_)
            button = _make_button(name, layout, display_function)

        buttons.append(button)
    if toggleable:
        display_function()
    result = ipywidgets.VBox([_no_resize(view), ipywidgets.HBox(buttons), slice_slider])
    result.update = display_function
    return result


def _make_button(name, layout, display_function):
    import ipywidgets
    if layout is None:
        button = ipywidgets.Button(description=name)
    else:
        button = ipywidgets.Button(description=name, layout=layout)

    def act(event=None):
        display_function()

    button.on_click(act)
    return button

def _make_toggle_button(name, layout, value, display_function):
    import ipywidgets
    if layout is None:
        button = ipywidgets.ToggleButton(description=name, value=value)
    else:
        button = ipywidgets.ToggleButton(description=name, layout=layout, value=value)

    def act(event=None):
        display_function()

    button.observe(act, 'value')
    return button



from ._slice_viewer import _SliceViewer

def interact(func,
             image = None,
             *args,
             continuous_update: bool = True,
             context:dict = None,
             zoom_factor:float = 1.0,
             zoom_spline_order:int = 0,
             colormap:str = None,
             display_min:float = None,
             display_max:float = None,
             viewer: _SliceViewer = None,
             **kwargs):
    """Takes a function which has an image as first parameter and additional parameters.
    It will build a user interface consisting of sliders for numeric parameters and parameters
    that are called "footprint" or "selem".

    Parameters
    ----------
    func : function
    image : Image, optional
        If not provided, context must be provided instead.
    args
    continuous_update : bool, optional
        Update the image while dragging the mouse, default: False
    context:dict, optional
        A dictionary of (name:image), allows showing a pulldown of available images, e.g.: globals()
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities
    viewer: _SliceViewer, optional
        The viewer where the result image should be shown.
    kwargs

    """
    import inspect
    import ipywidgets
    from ._utilities import parameter_is_image_parameter
    from ._context import Context
    from ._utilities import _no_resize
    from ._slice_viewer import _SliceViewer

    if 'cupy.ndarray' in str(type(image)):
        image = image.get()


    # hidden feature: func can be a tuple of (function, alias_name)
    if isinstance(func, tuple):
        func_name = func[1]
        func = func[0]
    else:
        func_name = func.__name__

    exposable_parameters = []
    footprint_parameters = []
    image_parameters = []

    if context is not None and not isinstance(context, Context):
        context = Context(context)

    image_passed = image is not None
    if context is not None and image is None:
        image = next(iter(context._images.values()))

    sig = inspect.signature(func)
    for key in sig.parameters.keys():
        exposable = False
        default_value = 0
        if isinstance(sig.parameters[key].default, int) or isinstance(sig.parameters[key].default, float) or isinstance(sig.parameters[key].default, str):
            default_value = sig.parameters[key].default
        min_value, max_value, step = guess_range(key, sig.parameters[key].annotation)

        #if min_value is not None:
        int_slider = ipywidgets.IntSlider
        float_slider = ipywidgets.FloatSlider
        #else:
        #    int_slider = ipywidgets.IntText
        #    float_slider = ipywidgets.FloatText
        if min_value is None:
            min_value = 0
        if max_value is None:
            max_value = 100

        if sig.parameters[key].annotation is str:
            default_value = ipywidgets.Text(value=default_value,
                                       continuous_update=continuous_update)
            exposable = True
        elif sig.parameters[key].annotation is int:
            default_value = int_slider(min=min_value, max=max_value, step=step, value=default_value, continuous_update=continuous_update)
            exposable = True
        elif sig.parameters[key].annotation is float or 'sigma' in key or 'radius' in key:
            default_value = float_slider(min=min_value, max=max_value, step=step, value=default_value, continuous_update=continuous_update)
            exposable = True
        elif key.startswith("is_") or sig.parameters[key].annotation is bool:
            default_value = ipywidgets.Checkbox(value=default_value)
            exposable = True
        elif key == 'footprint' or key == 'selem' or key == 'structuring_element':
            footprint_parameters.append(key)
            default_value = ipywidgets.IntSlider(min=min_value, max=max_value, step=step, value=default_value, continuous_update=continuous_update)
            exposable = True
        elif parameter_is_image_parameter(sig.parameters[key]) and "destination" not in key and key != "out"  and key != "output":
            if context is not None:
                image_parameters.append(key)
                default_value = ipywidgets.Dropdown(
                    options=list(context._images.keys())
                    #options=[(k, v) for k, v in context._images.items()],
                )
                exposable = True

        if exposable:
            if key in kwargs.keys():
                default_value = kwargs[key]
            exposable_parameters.append(inspect.Parameter(key, inspect.Parameter.KEYWORD_ONLY, default=default_value))

    viewer_was_none = viewer is None
    if viewer_was_none:
        viewer = _SliceViewer(image, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order, colormap=colormap, display_min=display_min, display_max=display_max)
    viewer.slice_slider.continuous_update=continuous_update
    command_label = ipywidgets.Label(value=func_name + "()")
    command_label.style.font_family = "Courier"

    from skimage import morphology
    execution_blocked = True

    def worker_function(*otherargs, **kwargs):

        command = func_name + "("
        if image_passed:
            command = command + "..."

        for key in [e.name for e in exposable_parameters]:

            if key in footprint_parameters:
                if len(image.shape) == 2:
                    command = command + ", " + key + "=disk(" + str(kwargs[key]) + ")"
                    kwargs[key] = morphology.disk(kwargs[key])
                elif len(image.shape) == 3:
                    command = command + ", " + key + "=ball(" + str(kwargs[key]) + ")"
                    kwargs[key] = morphology.ball(kwargs[key])
            elif key in image_parameters:
                command = command + ", " + key + "=" + str(kwargs[key])
                kwargs[key] = context._images[kwargs[key]]
            else:
                command = command + ", " + key + "=" + str(kwargs[key])
        command = command + ")"
        command_label.value = command.replace("(,", "(")

        if not execution_blocked:
            if image_passed:
                viewer.image = func(image, *args, **kwargs)
            else:
                viewer.image = func(*args, **kwargs)

        viewer.slice_slider.max = viewer.image.shape[0] - 1
        viewer.configuration_updated(None)

    worker_function.__signature__ = inspect.Signature(exposable_parameters)
    inter = ipywidgets.interactive(worker_function, dict(manual=False, auto_display=False))

    execution_blocked = False

    inter.update()

    output_widgets = []
    output_widgets.append(inter)
    output_widgets.append(command_label)

    if viewer_was_none:
        output_widgets.append(_no_resize(viewer.view))
        output_widgets.append(viewer.slice_slider)

    result = ipywidgets.VBox(output_widgets)
    result.update = viewer.configuration_updated
    return result


def guess_range(name, annotation):
    if name == 'footprint' or name == 'selem' or name == 'structuring_element':
        return 0, 100, 1
    if 'sigma' in name:
        return 0, 25, 1
    if 'radius' in name:
        return 0, 100, 1
    if 'factor' in name:
        return 0, 100, 1
    if name == 'angle' or "degrees" in name:
        return 0, 360, 15
    return None, None, None



import warnings

import numpy as np

class _SliceViewer():
    def __init__(self,
                 image,
                 slice_number: int = None,
                 axis: int = 0,
                 display_width: int = None,
                 display_height: int = None,
                 continuous_update: bool = True,
                 slider_text: str = "Slice",
                 zoom_factor:float = 1.0,
                 zoom_spline_order:int = 0,
                 colormap:str = None,
                 display_min:float = None,
                 display_max:float = None
                 ):
        import ipywidgets
        from ._image_widget import ImageWidget

        self.image = image
        self.axis = axis

        if slice_number is None:
            slice_number = int(image.shape[axis] / 2)

        if len(self.image.shape) > 2: # and self.image.shape[-1] != 3:
            sliced_image = np.take(image, slice_number, axis=axis)
        else:
            sliced_image = image
        self.view = ImageWidget(sliced_image,
                                zoom_factor=zoom_factor,
                                zoom_spline_order=zoom_spline_order,
                                colormap=colormap,
                                display_min=display_min,
                                display_max=display_max)

        # setup user interface for changing the slice
        self.slice_slider = ipywidgets.IntSlider(
            value=slice_number,
            min=0,
            max=image.shape[axis] - 1,
            continuous_update=continuous_update,
            description=slider_text,
        )
        # widgets.link((sliders1, 'value'), (slider2, 'value'))

        # connect user interface with event
        self.slice_slider.observe(self.update)

        self.update()

    # event handler when the user changed something:
    def update(self, event=None):
        if len(self.image.shape) == 3 and self.image.shape[-1] != 3:
            self.slice_slider.layout.display = None
            self.view.data = np.take(self.image, self.slice_slider.value, axis=self.axis)
        elif len(self.image.shape) == 4 and self.image.shape[-1] == 3:
            self.slice_slider.layout.display = None
            self.view.data = np.take(self.image, self.slice_slider.value, axis=self.axis)
        elif len(self.image.shape) == 4:
            raise NotImplementedError("Only 2D and 3D images are supported" + str(self.image.shape))
        else:
            self.view.data = self.image
            self.slice_slider.layout.display = 'none'

    def configuration_updated(self, event=None):
        warnings.warn('SliceViewer.configuration_updated is deprecated, use SliceViewer.update instead.')
        return self.update(event)




def display_range(
        image,
        slice_number: int = None,
        axis: int = 0,
        continuous_update: bool = True,
        zoom_factor :float = 1.0,
        zoom_spline_order :int = 0,
        colormap:str = None,
        display_min:float = None,
        display_max:float = None,
):
    """Shows an image and allows to change the display range using two sliders (min, max intensity).

    Parameters
    ----------
    image : image
        Image shown on the left (behind the curtain)
    slice_number : int, optional
        Slice-position in case we are looking at an image stack
    axis : int, optional
        Axis in case we are slicing a stack
    display_width : int, optional
        This parameter is obsolete. Use zoom_factor instead
    display_height : int, optional
        This parameter is obsolete. Use zoom_factor instead
    continuous_update : bool, optional
        Update the image while dragging the mouse, default: False
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities

    Returns
    -------
    An ipywidget with an image display and a slider.
    """
    import ipywidgets
    from ._image_widget import ImageWidget
    import numpy as np
    from ._utilities import _no_resize

    if 'cupy.ndarray' in str(type(image)):
        image = image.get()

    slice_slider = None

    if slice_number is None:
        slice_number = int(image.shape[axis] / 2)

    # setup user interface for changing the slice
    slice_slider = ipywidgets.IntSlider(
        value=slice_number,
        min=0,
        max=image.shape[axis ] -1,
        continuous_update=continuous_update,
        description="Slice"
    )
    if len(image.shape) < 3 or (len(image.shape) == 3 and image.shape[-1] == 3):
        slice_slider.layout.display = 'none'

    if display_min is None:
        display_min = np.min(image)
    if display_max is None:
        display_max = np.max(image)

    min_slider = ipywidgets.IntSlider(
        value=display_min,
        min=display_min,
        max=display_max,
        continuous_update=continuous_update,
        description="Minimum"
    )

    max_slider = ipywidgets.IntSlider(
        value=display_max,
        min=display_min,
        max=display_max,
        continuous_update=continuous_update,
        description="Maximum"
    )

    from ._image_widget import _img_to_rgb
    def transform_image():
        if len(image.shape) < 3 or (len(image.shape) == 3 and image.shape[-1] == 3):
            image_slice = _img_to_rgb(image.copy(), colormap=colormap, display_min=min_slider.value, display_max=max_slider.value)
        else:
            image_slice = _img_to_rgb(np.take(image, slice_slider.value, axis=axis), colormap=colormap, display_min=min_slider.value, display_max=max_slider.value)
        return image_slice
    view = ImageWidget(transform_image(), zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order)

    # event handler when the user changed something:
    def configuration_updated(event=None):
        view.data = transform_image()

    configuration_updated(None)

    # connect user interface with event
    slice_slider.observe(configuration_updated)
    min_slider.observe(configuration_updated)
    max_slider.observe(configuration_updated)

    result = ipywidgets.VBox([_no_resize(view), slice_slider, min_slider, max_slider])
    result.update = configuration_updated
    return result



import numpy as np
def merge_rgb(image_red, image_green, image_blue):
    """
    Turns three images (channels) into a scikit-image-style RGB image
    with the last dimension being the channel axis.
    """
    return np.asarray([image_red, image_green, image_blue]).swapaxes(0, 2)

def is_image(image):
    return isinstance(image, np.ndarray) or \
        str(type(image)) in ["<class 'cupy._core.core.ndarray'>",
                             "<class 'dask.array.core.Array'>",
                             "<class 'xarray.core.dataarray.DataArray'>",
                             "<class 'resource_backed_dask_array.ResourceBackedDaskArray'>",
                             "<class 'torch.Tensor'>",
                             "<class 'pyclesperanto_prototype._tier0._cuda_backend.CUDAArray'>",
                             "<class 'pyclesperanto_prototype._tier0._pycl.OCLArray'>"]
        # isinstance(image, tuple) or \
        # isinstance(image, list) or \


def is_label_image(image):
    if isinstance(image, tuple) or \
            isinstance(image, list):
        return False

    return image.dtype == np.uint32 or image.dtype == np.uint64 or \
        image.dtype == np.int32 or image.dtype == np.int64


def count_image_parameters(func):
    import inspect
    try:
        sig = inspect.signature(func)
    except:
        return 0

    n = 0

    for i, key in enumerate(list(sig.parameters.keys())):
        if parameter_is_image_parameter(sig.parameters[key]):
            n = n + 1
    return n


def parameter_is_image_parameter(parameter):
    type_annotation = str(parameter.annotation)
    name = parameter.name

    return ("NewType.<locals>.new_type" in type_annotation or \
            "Image" in type_annotation or \
            "LabelsData" in type_annotation or \
            "LayerData" in type_annotation or \
            "image" in name.lower() or \
            "label" in name.lower() or \
            "mask" in name.lower()
            )

def _no_resize(widget):
    import ipywidgets
    return ipywidgets.HBox([ipywidgets.VBox([widget])])


logo = np.asarray([
        [3,3,3,3,3,3,3,3,3,3,3],
        [3,3,3,3,3,3,3,3,3,3,3],
        [3,3,3,3,3,3,3,3,3,3,3],
        [3,3,4,4,4,2,3,3,3,2,3],
        [3,4,3,3,3,2,3,3,3,2,3],
        [3,3,4,4,3,3,2,3,2,3,3],
        [3,3,3,3,4,3,2,3,2,3,3],
        [3,4,4,4,3,3,3,2,3,3,3],
        [3,3,3,3,3,3,3,3,3,3,3],
        [3,3,3,3,3,3,3,3,3,3,3],
        [3,3,3,3,3,3,3,3,3,3,3],
    ]).astype(np.uint32)


def numpy_to_gif_bytestream(timelapse, frame_delay_ms=100, num_loops=1000):
    """Turn a NumPy array into a bytestream"""
    import numpy as np
    import imageio
    import io

    # Convert the NumPy array to a PIL Image
    # image = Image.fromarray(data.astype(np.uint8)).convert("RGBA")

    # Create a BytesIO object
    bytes_io = io.BytesIO()

    # Save the PIL image to the BytesIO object as a PNG
    # image.save(bytes_io, format='GIF')

    with imageio.get_writer(bytes_io, mode='I', duration=frame_delay_ms, loop=num_loops, format="GIF") as writer:
        for frame in timelapse:
            writer.append_data(frame.astype(np.uint8))

    # return the beginning of the file as a bytestream
    bytes_io.seek(0)
    return bytes_io.read()


def _gif_to_html(gif, width=None):
    import base64

    style = ""
    if width is not None:
        style = f"style=\"width: {width}px;\""

    url = 'data:image/gif;base64,' + base64.b64encode(gif).decode('utf-8')
    return f'<img src="{url}" {style}></img>'



from ipycanvas import Canvas
import numpy as np
from ._colormaps import _labels_lut # for internal backwards compatibility

class ImageWidget(Canvas):
    def __init__(self, image, zoom_factor:float=1.0, zoom_spline_order:int=0, colormap:str=None, display_min:float=None, display_max:float=None):
        if not ((len(image.shape) == 2) or (len(image.shape) in [3, 4] and image.shape[-1] == 3)):
            raise NotImplementedError("Only 2D images are supported" + str(image.shape))
        height = image.shape[0] * zoom_factor
        width = image.shape[1] * zoom_factor
        self.zoom_factor = zoom_factor
        self.zoom_spline_order = zoom_spline_order
        super().__init__(width=width * zoom_factor, height=height * zoom_factor)
        self.colormap = colormap
        self.display_min = display_min
        self.display_max = display_max
        self.data = np.asarray(image)
        self.layout.stretch = False

    @property
    def data(self):
        """Image data as numpy array
        """
        return self._data

    @data.setter
    def data(self, new_data):
        """Take in new image data, compress to PNG, send to image widget.
        """
        if new_data is None:
            return

        self._data = np.asarray(new_data)
        self._update_image()
        self.height = self._data.shape[0] * self.zoom_factor
        self.width = self._data.shape[1] * self.zoom_factor

    def _update_image(self):
        if self.zoom_factor == 1.0:
            self.put_image_data(_img_to_rgb(self._data, colormap=self.colormap, display_min=self.display_min, display_max=self.display_max), 0, 0)
        else:
            zoomed = self._zoom(self._data)
            self.put_image_data(_img_to_rgb(zoomed, colormap=self.colormap, display_min=self.display_min, display_max=self.display_max), 0, 0)

    def _zoom(self, data):
        if len(data.shape) > 2 and data.shape[-1] == 3:
            # handle RGB images
            return np.asarray([self._zoom(data[:,:,i]) for i in range(data.shape[2])]).swapaxes(0, 2).swapaxes(1, 0)

        from scipy.ndimage import affine_transform
        matrix = np.asarray([[1.0 / self.zoom_factor, 0, -0.5],
                             [0, 1.0 / self.zoom_factor, -0.5],
                             [0, 0, 1],
                             ])
        zoomed_shape = (np.asarray(data.shape) * self.zoom_factor).astype(int)
        zoomed = affine_transform(data,
                                  matrix,
                                  output_shape=zoomed_shape,
                                  order=self.zoom_spline_order,
                                  mode='nearest')
        return zoomed


def _is_label_image(image):
    return image.dtype == np.uint32 or image.dtype == np.uint64 or \
           image.dtype == np.int32 or image.dtype == np.int64


def _img_to_rgb(image,
                colormap=None,
                display_min=None,
                display_max=None):
    from ._colormaps import _labels_lut, create_colormap

    if len(image.shape) > 2 and image.shape[-1] == 3:
        return image

    if image.dtype == bool:
        image = image * 1

    if _is_label_image(image):
        lut = _labels_lut()
        return np.asarray([lut[:, c].take(image.astype(np.int64)) for c in range(0, 3)]).swapaxes(0, 2).swapaxes(1, 0) * 255

    if display_min is None:
        display_min = image.min()
    if display_max is None:
        display_max = image.max()

    diplay_range_width = (display_max - display_min)
    if diplay_range_width == 0:
        diplay_range_width = 1

    image = (image.astype(float) - display_min) / diplay_range_width * 255
    image = np.minimum(image, 255)
    image = np.maximum(image, 0)

    if colormap is None:
        return np.asarray([image, image, image]).swapaxes(0, 2).swapaxes(1, 0)
    else:
        lut = np.asarray(create_colormap(colormap).colors)
        return np.asarray([lut[:, c].take(image.astype(int)) for c in range(0, 3)]).swapaxes(0, 2).swapaxes(1, 0) * 255



# the code in this file is modified from
# https://github.com/guiwitz/microfilm/blob/c0666e19b77db66b0af3a4d3759baaf19243b9db/microfilm/colorify.py#L11

# BSD 3-Clause License
#
# Copyright (c) 2021, Bern University, Mathematical Institute and Microscopy
# Imaging Center, Guillaume Witz
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from functools import lru_cache

@lru_cache(maxsize=16)
def create_colormap(cmap_name, num_colors=256):
    """
    Return a colormap defined by its name

    Parameters
    ----------
    cmap_name: str
        {'pure_red', 'pure_green', 'pure_blue', 'pure_magenta',
        'pure_cyan', 'pure_yellow'} or Matplotlib colormap
        Colors will be reversed if the name ends with '_r'
    num_colors: int
        number of steps in color scale
    Returns
    -------
    cmap: Matplotlib colormap
    """
    import matplotlib.pyplot as plt
    import numpy as np
    from matplotlib.colors import ListedColormap

    if cmap_name in plt.colormaps():
        flip_map = False
        cmap = _convert_to_listed_colormap(plt.get_cmap(cmap_name, num_colors), num_colors)
    else:
        flip_map = cmap_name.endswith("_r")
        if flip_map:
            cmap_name = cmap_name[:-2]

        if cmap_name == 'pure_red':
            cmap = ListedColormap(np.c_[np.linspace(0, 1, num_colors), np.zeros(num_colors), np.zeros(num_colors)])
        elif cmap_name == 'pure_green':
            cmap = ListedColormap(np.c_[np.zeros(num_colors), np.linspace(0, 1, num_colors), np.zeros(num_colors)])
        elif cmap_name == 'pure_blue':
            cmap = ListedColormap(np.c_[np.zeros(num_colors), np.zeros(num_colors), np.linspace(0, 1, num_colors)])
        elif cmap_name == 'pure_cyan':
            cmap = ListedColormap(np.c_[np.zeros(num_colors), np.linspace(0, 1, num_colors), np.linspace(0, 1, num_colors)])
        elif cmap_name == 'pure_magenta':
            cmap = ListedColormap(np.c_[np.linspace(0, 1, num_colors), np.zeros(num_colors), np.linspace(0, 1, num_colors)])
        elif cmap_name == 'pure_yellow':
            cmap = ListedColormap(np.c_[np.linspace(0, 1, num_colors), np.linspace(0, 1, num_colors), np.zeros(num_colors)])
        elif cmap_name == 'segmentation':
            cmap = ListedColormap(_labels_lut())
        elif cmap_name == 'ran_gradient':
            cmap = random_grandient_cmap(num_colors=num_colors)
        else:
            raise Exception(f"Your colormap {cmap_name} doesn't exist either in Matplotlib or microfilm.")

    if flip_map:
        cmap = cmap.reversed()

    return cmap


def random_grandient_cmap(num_colors=25, seed=42):
    """Create a colormap as the gradient of a given random color"""
    from skimage.color import hsv2rgb
    import numpy as np
    from matplotlib.colors import ListedColormap

    rgb = hsv2rgb([np.random.random(1)[0], 0.95, 0.95])

    cmap = ListedColormap(np.c_[
        np.linspace(0,rgb[0],num_colors),
        np.linspace(0,rgb[1],num_colors),
        np.linspace(0,rgb[2],num_colors)
        ])
    return cmap


@lru_cache(maxsize=1)
def _labels_lut():
    from numpy.random import MT19937
    from numpy.random import RandomState, SeedSequence
    rs = RandomState(MT19937(SeedSequence(3)))
    lut = rs.rand(65537, 3)
    lut[0, :] = 0
    # these are the first four colours from matplotlib's default
    lut[1] = [0.12156862745098039, 0.4666666666666667, 0.7058823529411765]
    lut[2] = [1.0, 0.4980392156862745, 0.054901960784313725]
    lut[3] = [0.17254901960784313, 0.6274509803921569, 0.17254901960784313]
    lut[4] = [0.8392156862745098, 0.15294117647058825, 0.1568627450980392]
    return lut


def _convert_to_listed_colormap(colormap, num_samples):
    import numpy as np
    from matplotlib.colors import ListedColormap
    samples = np.linspace(0, 1, num_samples)
    colors = colormap(samples)
    return ListedColormap(colors)



import warnings


def imshow(image,
           title:str = None,
           labels: bool = None,
           min_display_intensity: float = None,
           max_display_intensity: float = None,
           plot=None,
           colorbar: bool = False,
           colormap=None,
           alpha: float = None,
           continue_drawing: bool = False,
           axes:bool = False,
           cmap:str = None,
           vmin:float = None,
           vmax:float = None):
    """Visualize an image, e.g. in Jupyter notebooks.

    Parameters
    ----------
    image: np.ndarray
        numpy or OpenCL-backed image to visualize
    title: str

    labels: bool
        None: auto-detect (default): if image of type uint16, 32 or 64: True, False otherwise
        True: integer labels will be visualized with colors
        False: Specified or default colormap will be used to display intensities.
    min_display_intensity: float
        lower limit for display range
    max_display_intensity: float
        upper limit for display range
    plot: matplotlib axis
        Plot object where the image should be shown. Useful for putting multiple images in subfigures.
    colorbar: bool
        True puts a colorbar next to the image. Will not work with label images and when visualizing multiple
        images (continue_drawing=True).
        Fqlse: no colorbar (default)
    colormap: str
        matplotlib colormap or microfilm names such as "pure_green", "pure_magenta",...
    alpha: float
        blending value (between 0 and 1)
    continue_drawing: bool, optional
        True: the next shown image can be visualized on top of the current one, e.g. with alpha = 0.5
        False: the image is shown instantly and the cache is emptied (default)
    axes: bool
        turn axes on/off (default: False)
    """
    import numpy as np
    from ._utilities import is_label_image
    from ._colormaps import _labels_lut, create_colormap

    if cmap is not None:
        warnings.warn("The parameter cmap is deprecated, use colormap instead.")
        colormap = cmap

    if vmin is not None:
        warnings.warn("The parameter min_display_intensity is deprecated, use min_display_intensity instead.")
        min_display_intensity = vmin

    if vmax is not None:
        warnings.warn("The parameter max_display_intensity is deprecated, use max_display_intensity instead.")
        max_display_intensity = vmax

    if len(image.shape) == 3 and image.shape[-1] > 4:
        image = image.max(axis=0)

    if 'cupy.ndarray' in str(type(image)):
        image = image.get()

    image = np.asarray(image)
    if len(image.shape) == 1:
        image = image[np.newaxis]

    if labels is None:
        labels = is_label_image(image)

    if colormap is None:
        colormap = "Greys_r"

    if colormap.startswith("pure_"):
        colormap = create_colormap(colormap)

    if labels:
        import matplotlib
        import numpy as np

        if not hasattr(imshow, "labels_cmap"):
            lut = _labels_lut()
            imshow.labels_cmap = matplotlib.colors.ListedColormap(lut)
        colormap = imshow.labels_cmap

        if min_display_intensity is None:
            min_display_intensity = 0
        if max_display_intensity is None:
            max_display_intensity = 65536

    if plot is None:
        import matplotlib.pyplot as plt
        plt.imshow(image, cmap=colormap, vmin=min_display_intensity, vmax=max_display_intensity, interpolation='nearest',
                   alpha=alpha)

        if axes is not None:
            plt.axis('on' if axes else 'off')

        if title is not None:
            plt.title(title)

        if colorbar:
            plt.colorbar()
        if not continue_drawing:
            plt.show()
    else:
        ims = plot.imshow(image, cmap=colormap, vmin=min_display_intensity, vmax=max_display_intensity,
                          interpolation='nearest', alpha=alpha)
        if colorbar:
            fig = plot.get_figure()
            cax = fig.add_axes([plot.get_position().x1 + 0.01, plot.get_position().y0, 0.005,
                                plot.get_position().height])
            fig.colorbar(ims, cax=cax)

        if title is not None:
            plot.set_title(title)

        if axes is not None:
            plot.axis('on' if axes else 'off')



def slice(
        image,
        slice_number : int = None,
        axis : int = 0,
        display_width : int = None,
        display_height : int = None,
        continuous_update:bool = True,
        slider_text:str="Slice",
        zoom_factor:float = 1.0,
        zoom_spline_order:int = 0,
        colormap:str = None,
        display_min:float = None,
        display_max:float = None
):
    """Shows an image with a slider to go through a stack.

    Parameters
    ----------
    image : image
        Image shown
    slice_number : int, optional
        Slice-position in the stack
    axis : int, optional
        Axis in case we are slicing a stack
    display_width : int, optional
        This parameter is obsolete. Use zoom_factor instead
    display_height : int, optional
        This parameter is obsolete. Use zoom_factor instead
    continuous_update : bool, optional
        Update the image while dragging the mouse, default: False
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities

    Returns
    -------
    An ipywidget with an image display and a slider.
    """
    from ._utilities import _no_resize
    from ._slice_viewer import _SliceViewer

    if 'cupy.ndarray' in str(type(image)):
        image = image.get()

    import ipywidgets
    viewer = _SliceViewer(image,
        slice_number,
        axis,
        display_width,
        display_height,
        continuous_update,
        slider_text,
        zoom_factor=zoom_factor,
        zoom_spline_order=zoom_spline_order,
        colormap=colormap,
        display_min=display_min,
        display_max=display_max
    )
    view = viewer.view
    slice_slider = viewer.slice_slider

    result = ipywidgets.VBox([_no_resize(view), slice_slider])
    result.update = viewer.update
    return result



import numpy as np

blacklist = ["try_all_threshold",
             "test",
             "minimum_of_touching_neighbors_map",
             "maximum_of_touching_neighbors_map",
             "standard_deviation_of_touching_neighbors_map",
             "mean_of_touching_neighbors_map",
             "z_position_of_minimum_z_projection",
             "z_position_of_maximum_z_projection",
             ]

def nop(image):
    """No operation, returns the image unchanged"""
    return image



class Context():
    def __init__(self, variables):
        from ._utilities import logo

        self._functions = {"stackview.nop": nop}
        self._modules = { }
        self._images = {"no_image": logo}
        #self._label_images = {}

        self.parse(variables.items())

    def parse(self, items, prefix: str = None):

        from types import ModuleType
        from typing import Callable
        import stackview
        from ._utilities import is_image, is_label_image, count_image_parameters

        for name, value in items:
            if not name.startswith("_"):
                if isinstance(value, ModuleType) and value is not stackview:
                    if prefix is None:
                        self._modules[name] = value
                        self.parse({key: getattr(value, key) for key in dir(value)}.items(), prefix=name + ".")
                elif isinstance(value, Callable):
                    if count_image_parameters(value) > 0:
                        if name not in blacklist:
                            if prefix is None:
                                self._functions[name] = value
                            else:
                                self._functions[prefix + name] = value
                elif is_image(value):
                    if prefix is None:
                        #if is_label_image(value):
                        #    self._label_images[name] = value
                        #else:
                            self._images[name] = value




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="stackview",
    version="0.7.10",
    author="Robert Haase",
    author_email="robert.haase@uni-leipzig.de",
    description="Interactive image stack viewing in jupyter notebooks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/haesleinhuepf/stackview/",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=["numpy!=1.19.4", "ipycanvas", "ipywidgets", "scikit-image", "ipyevents", "toolz", "matplotlib", "ipykernel", "imageio"],
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha",
    ],
    entry_points={
        'bia_bob_plugins': [
            'stackview_plugins = stackview._bia_bob_plugins:list_bia_bob_plugins',
        ]
    }
)



import warnings

def side_by_side(
        image1,
        image2,
        slice_number: int = None,
        axis: int = 0,
        display_width: int = None,
        display_height: int = None,
        continuous_update: bool = True,
        slider_text: str = "Slice",
        zoom_factor:float = 1.0,
        zoom_spline_order:int = 0,
        colormap:str = None,
        display_min:float = None,
        display_max:float = None
):
    """Shows two images in magenta and green plus a third with their colocalization / overlap view and
    a slider to go through a stack.

    Parameters
    ----------
    image1 : image
        Image shown on the left
    image2 : image
        Image shown on the right
    slice_number : int, optional
        Slice-position in the stack
    axis : int, optional
        Axis in case we are slicing a stack
    display_width : int, optional
        This parameter is obsolete. Use zoom_factor instead
    display_height : int, optional
        This parameter is obsolete. Use zoom_factor instead
    continuous_update : bool, optional
        Update the image while dragging the mouse, default: False
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities

    Returns
    -------
    An ipywidget with three image displays and a slider.
    """

    import ipywidgets
    from ._image_widget import ImageWidget
    import numpy as np
    from ._utilities import _no_resize

    if 'cupy.ndarray' in str(type(image1)):
        image1 = image1.get()

    if 'cupy.ndarray' in str(type(image2)):
        image2 = image2.get()

    if slice_number is None:
        slice_number = int(image1.shape[axis] / 2)

    if len(image1.shape) <= 2:
        slice_image = image1
    else:
        slice_image = np.take(image1, slice_number, axis=axis)

    zeros_image = np.zeros(slice_image.shape)
    view1 = ImageWidget(slice_image, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order)
    view2 = ImageWidget(slice_image, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order)
    view3 = ImageWidget(slice_image, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order)

    # setup user interface for changing the slice
    slice_slider = None
    if len(image1.shape) > 2:
        slice_slider = ipywidgets.IntSlider(
            value=slice_number,
            min=0,
            max=image1.shape[0] - 1,
            continuous_update=continuous_update,
            description=slider_text,
        )

    from ._image_widget import _is_label_image, _img_to_rgb

    # event handler when the user changed something:
    def configuration_updated(event=None):
        if slice_slider is not None:
            z = slice_slider.value
            slice_image1 = np.take(image1, z, axis=axis)
            slice_image2 = np.take(image2, z, axis=axis)
        else:
            slice_image1 = image1
            slice_image2 = image2

        if _is_label_image(slice_image1) or _is_label_image(slice_image2):
            rgb_image1 = _img_to_rgb(slice_image1, colormap=colormap, display_min=display_min, display_max=display_max)
            rgb_image2 = _img_to_rgb(slice_image2, colormap=colormap, display_min=display_min, display_max=display_max)

            if _is_label_image(slice_image1) and _is_label_image(slice_image2):
                warnings.warn("Side-by-side mixing two label images may look weird." +
                              "Consider showing original image and a label image side-by-side.")
                factor1 = 0.5
                factor2 = 0.5
            elif _is_label_image(slice_image1):
                factor1 = 0.3
                factor2 = 0.7
            elif _is_label_image(slice_image2):
                factor1 = 0.7
                factor2 = 0.3

            rgb_mix = factor1 * rgb_image1 + factor2 * rgb_image2

            view1.data = rgb_image1
            view2.data = rgb_image2
            view3.data = rgb_mix
        else:
            view1.data = np.asarray([slice_image1, zeros_image, slice_image1]).swapaxes(0, 2)
            view2.data = np.asarray([zeros_image, slice_image2, zeros_image]).swapaxes(0, 2)
            view3.data = np.asarray([slice_image1, slice_image2, slice_image1]).swapaxes(0, 2)

    configuration_updated(None)

    if slice_slider is not None:
        # connect user interface with event
        slice_slider.observe(configuration_updated)

        result = ipywidgets.VBox([
            ipywidgets.HBox([_no_resize(view1), _no_resize(view2), _no_resize(view3)]),
            slice_slider
        ])
    else:
        result = ipywidgets.HBox([_no_resize(view1), _no_resize(view2), _no_resize(view3)])

    result.update = configuration_updated
    return result




def orthogonal(
        image,
        display_width : int = None,
        display_height : int = None,
        continuous_update:bool=True,
        zoom_factor:float = 1.0,
        zoom_spline_order:int = 0,
        colormap:str = None,
        display_min:float = None,
        display_max:float = None
):
    """Show three viewers slicing the image stack in Z,Y and X.

    Parameters
    ----------
    image : image
        Image to be displayed
    display_width : int, optional
        This parameter is obsolete. Use zoom_factor instead
    display_height : int, optional
        This parameter is obsolete. Use zoom_factor instead
    continuous_update : bool, optional
        Update the image while dragging the mouse, default: False
    zoom_factor: float, optional
        Allows showing the image larger (> 1) or smaller (<1)
    zoom_spline_order: int, optional
        Spline order used for interpolation (default=0, nearest-neighbor)
    colormap: str, optional
        Matplotlib colormap name or "pure_green", "pure_magenta", ...
    display_min: float, optional
        Lower bound of properly shown intensities
    display_max: float, optional
        Upper bound of properly shown intensities

    Returns
    -------
    An ipywidget with an image display and a slider.

    See Also
    --------
    slice()
    """
    import ipywidgets
    from ._slice import slice

    if 'cupy.ndarray' in str(type(image)):
        image = image.get()

    widgets = [
        slice(image, axis=0, slider_text="Z", continuous_update=continuous_update, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order, colormap=colormap, display_min=display_min, display_max=display_max),
        slice(image, axis=1, slider_text="Y", continuous_update=continuous_update, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order, colormap=colormap, display_min=display_min, display_max=display_max),
        slice(image, axis=2, slider_text="X", continuous_update=continuous_update, zoom_factor=zoom_factor, zoom_spline_order=zoom_spline_order, colormap=colormap, display_min=display_min, display_max=display_max),
    ]

    def update(event=None):
        for widget in widgets:
            widget.update()

    result = ipywidgets.HBox(widgets)
    result.update = update
    return result
